/*****************************************************************
OpenCL parallel integrator based on Runge-Kutta schemes.

Copyright 2014 Marcos Rodriguez, Fernando Blesa and Roberto Barrio
http://cody.unizar.es/software/
Contact: <marcos@unizar.es>

This code is released under GPL license. See the web site above for citation instructions

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef __APPLE__
	#include <OpenCL/cl.h>
#else
	#include <CL/cl.h>
#endif


#include "parameters.h"


void fileToSource (char *filename, char **source);
real_t getRand ();

int main () {
	long int i, j;

	// RANDOM NUMBERS
	real_t *randomNumber;
	randomNumber = (real_t*) malloc (NVAR*N * sizeof (real_t));
	for (i=0; i<NVAR*N; i++)
		randomNumber[i] = getRand ();
	for (i=0; i<NVAR; i++) 
		randomNumber[N*i] = 0.0;

	// GLOBAL SIZES
	size_t global_work_size[] = {N,1};

	// READ KERNEL FILE
	char *source;
	fileToSource ("test_kernel.cl", &source);
	
	// OPENCL START AND BUILD KERNEL
	#include "openCL_start.h"
	#include "openCL_kernel_build.h"

	// ALLOCATE MEMORY FOR VARIABLES ON DEVICE
	cl_mem d_x = clCreateBuffer (context, CL_MEM_WRITE_ONLY, NVAR*N * sizeof (real_t), NULL, NULL);
	cl_mem d_rand = clCreateBuffer (context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, NVAR*N * sizeof (real_t), randomNumber, NULL);


	// SET KERNEL ARGUMENTS
	clSetKernelArg (testKernel, 0, sizeof (d_x), (void*) &d_x);
	clSetKernelArg (testKernel, 1, sizeof (d_rand), (void*) &d_rand);

	// EVENT FOR MEASURING TIME
	cl_event prof_event;

	// LAUNCH KERNEL!
	clEnqueueNDRangeKernel (queue,	// QUEUE
		testKernel,	// KERNEL NAME
		1,		// DIMENSIONS
		NULL,		// OFFSET
		global_work_size,// GLOBAL SIZE
		NULL,		// LOCAL SIZE
		0,		// NUM EVENTS IN WAIT LIST
		NULL,		// WAIT LIST
		&prof_event	// EVENT FOR MEASURING TIME 
	);

	// WAIT UNTIL ALL IS FINISHED
	clFinish (queue);
	clWaitForEvents (1, &prof_event);


	// READ VARIABLE X
	real_t *x = (real_t*) malloc (NVAR*N * sizeof (real_t));
	clEnqueueReadBuffer (queue, d_x, CL_TRUE, 0, NVAR*N * sizeof (real_t), x, 0, NULL, NULL);

	
	// PRINT VARIABLE X (FIRST THREAD)
	/*for (i=0; i<NVAR; i++) 
		printf ("x[%1li] = %.7le\n", i, x[N*i]);*/

	// CPU -- GPU time
	cl_ulong c0, c1;
	clGetEventProfilingInfo(prof_event, CL_PROFILING_COMMAND_START, sizeof(c0), &c0, NULL);
	clGetEventProfilingInfo(prof_event, CL_PROFILING_COMMAND_END, sizeof(c1), &c1, NULL);
	//printf("\nExecution time in milliseconds = %0.3f ms\n", ((real_t) (c1 - c0) / 1.e6));

	printf ("%0.3f\t\t\t%.7le ", ((real_t) (c1 - c0) / 1.e6), x[0]);
	

	// FREE MEMORY OBJECTS FROM HOST
	free (x);
	free (randomNumber);

	// FREE MEMORY OBJECTS FROM DEVICE
	clReleaseMemObject (d_x);
	clReleaseMemObject (d_rand);

	// FREE OPENCL OBJECTS
	clReleaseKernel (testKernel);
	clReleaseProgram (program);
	clReleaseCommandQueue (queue);
	clReleaseContext (context);

	return 0;
}




real_t getRand () {
	return (((double) rand()) / RAND_MAX) * 0.02 - 0.01;
}

void fileToSource (char *filename, char **source) {
	FILE* finp;

	finp = fopen (filename, "r");
	size_t fileLength;

	fseek(finp, 0, SEEK_END);
	fileLength = ftell(finp);
	rewind (finp);
	
	*source = (char*) malloc (fileLength*sizeof (char));
	int i=0;
	while (!feof (finp)) {
		EOF == fscanf (finp, "%c", &(*source)[i++]);
	}
	(*source)[fileLength-1] = '\0';	

	
	fclose (finp);
}

